@@ -8,12 +8,16 @@ Copyright (c) 2017-2022 Matej Aleksandrov, Žiga Lukšič (Sinergise)
 This source code is licensed under the MIT license found in the LICENSE
 file in the root directory of this source tree.
 """
+from __future__ import annotations
+
+from typing import Callable, List, Optional, Union, cast
 
 import numpy as np
 from sklearn.cluster import AgglomerativeClustering
 from sklearn.feature_extraction.image import grid_to_graph
 
-from eolearn.core import EOTask, FeatureType
+from eolearn.core import EOPatch, EOTask, FeatureType
+from eolearn.core.types import FeatureSpec, Literal
 
 
 class ClusteringTask(EOTask):
@@ -27,46 +31,37 @@ class ClusteringTask(EOTask):
 
     def __init__(
         self,
-        features,
-        new_feature_name,
-        distance_threshold=None,
-        n_clusters=None,
-        affinity="cosine",
-        linkage="single",
-        remove_small=0,
-        connectivity=None,
-        mask_name=None,
+        features: FeatureSpec,
+        new_feature_name: str,
+        distance_threshold: Optional[float] = None,
+        n_clusters: Optional[int] = None,
+        affinity: Literal["euclidean", "l1", "l2", "manhattan", "cosine"] = "cosine",
+        linkage: Literal["ward", "complete", "average", "single"] = "single",
+        remove_small: int = 0,
+        connectivity: Union[None, np.ndarray, Callable] = None,
+        mask_name: Optional[str] = None,
     ):
         """Class constructor
 
         :param features: A collection of features used for clustering. The features need to be of type DATA_TIMELESS
-        :type features: dict(FeatureType.DATA_TIMELESS: set(str))
         :param new_feature_name: Name of feature that is the result of clustering
-        :type new_feature_name: str
         :param distance_threshold: The linkage distance threshold above which, clusters will not be merged. If non None,
             n_clusters must be None nd compute_full_tree must be True
-        :type distance_threshold: float or None
         :param n_clusters: The number of clusters found by the algorithm. If distance_threshold=None, it will be equal
             to the given n_clusters
-        :type n_clusters: int or None
         :param affinity: Metric used to compute the linkage. Can be “euclidean”, “l1”, “l2”, “manhattan”, “cosine”.
-        :type affinity: str
         :param linkage: Which linkage criterion to use. The linkage criterion determines which distance to use between
             sets of observation. The algorithm will merge the pairs of cluster that minimize this criterion.
             - ward minimizes the variance of the clusters being merged.
             - average uses the average of the distances of each observation of the two sets.
             - complete or maximum linkage uses the maximum distances between all observations of the two sets.
             - single uses the minimum of the distances between all observations of the two sets.
-        :type linkage: {“ward”, “complete”, “average”, “single”}
         :param remove_small: If greater than 0, removes all clusters that have fewer points as "remove_small"
-        :type remove_small: int
         :param connectivity: Connectivity matrix. Defines for each sample the neighboring samples following a given
             structure of the data. This can be a connectivity matrix itself or a callable that transforms the data into
             a connectivity matrix, such as derived from neighbors_graph. If set to None it uses the graph that has
             adjacent pixels connected.
-        :type connectivity: array-like, callable or None
         :param mask_name: An optional mask feature used for exclusion of the area from clustering
-        :type mask_name: str
         """
         self.features_parser = self.get_feature_parser(features)
         self.distance_threshold = distance_threshold
@@ -74,7 +69,7 @@ class ClusteringTask(EOTask):
         self.linkage = linkage
         self.new_feature_name = new_feature_name
         self.n_clusters = n_clusters
-        self.compute_full_tree = "auto"
+        self.compute_full_tree: Union[Literal["auto"], bool] = "auto"
         if distance_threshold is not None:
             self.compute_full_tree = True
         if remove_small < 0:
@@ -83,29 +78,13 @@ class ClusteringTask(EOTask):
         self.connectivity = connectivity
         self.mask_name = mask_name
 
-    def construct_data(self, eopatch):
-        """
-        :param eopatch: EOPatch with all features that will be used
-        :return: array of vectors constructed from the features listed
-        """
-
-        data = None
-        for feature in self.features_parser.get_features(eopatch):
-            if data is None:
-                data = eopatch[feature]
-            else:
-                data = np.concatenate((data, eopatch[feature]), axis=2)
-
-        return data
-
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """
         :param eopatch: Input EOPatch
-        :type eopatch: EOPatch
         :return: Transformed EOPatch
-        :rtype: EOPatch
         """
-        data = self.construct_data(eopatch)
+        relevant_features = self.features_parser.get_features(eopatch)
+        data = np.concatenate([eopatch[feature] for feature in relevant_features], axis=2)
 
         # Reshapes the data, because AgglomerativeClustering method only takes one dimensional arrays of vectors
         org_shape = data.shape
@@ -150,8 +129,9 @@ class ClusteringTask(EOTask):
 
         # Transforms data back to original shape and setting all masked regions to -1
         if self.mask_name is not None:
+            locations = cast(List[int], locations)  # set because mask_name is not None
             new_data = [-1] * org_length
-            for i, val in zip(np.delete(range(org_length), locations), trimmed_labels):
+            for i, val in zip(np.delete(np.arange(org_length), locations), trimmed_labels):
                 new_data[i] = val
             trimmed_labels = new_data
 
