@@ -8,11 +8,15 @@ Copyright (c) 2017-2022 Matej Aleksandrov, Matic Lubej, Devis Peressutti, Žiga
 This source code is licensed under the MIT license found in the LICENSE
 file in the root directory of this source tree.
 """
-from abc import ABCMeta
+from __future__ import annotations
+
+from abc import ABCMeta, abstractmethod
+from typing import Optional
 
 import numpy as np
 
-from eolearn.core import EOTask
+from eolearn.core import EOPatch, EOTask
+from eolearn.core.types import SingleFeatureSpec
 
 
 class ReferenceScenesTask(EOTask):
@@ -23,31 +27,33 @@ class ReferenceScenesTask(EOTask):
     Contributor: Johannes Schmid, GeoVille Information Systems GmbH, 2018
 
     :param feature: Name of the eopatch data layer. Needs to be of the FeatureType "DATA".
-    :type feature: (FeatureType, str) or (FeatureType, str, str)
     :param valid_fraction_feature: Name of the layer containing the valid fraction obtained with the EOTask
-                                    'AddValidDataFraction'. Needs to be of the FeatureType "SCALAR".
-    :type valid_fraction_feature: (FeatureType, str)
+        'AddValidDataFraction'. Needs to be of the FeatureType "SCALAR".
     :param max_scene_number: Maximum number of reference scenes taken for the creation of the composite. By default,
-                            the maximum number of scenes equals the number of time frames
-    :type max_scene_number: int
+        the maximum number of scenes equals the number of time frames
 
     """
 
-    def __init__(self, feature, valid_fraction_feature, max_scene_number=None):
+    def __init__(
+        self,
+        feature: SingleFeatureSpec,
+        valid_fraction_feature: SingleFeatureSpec,
+        max_scene_number: Optional[int] = None,
+    ):
         self.renamed_feature = self.parse_renamed_feature(feature)
         self.valid_fraction_feature = self.parse_feature(valid_fraction_feature)
         self.number = max_scene_number
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         feature_type, feature_name, new_feature_name = self.renamed_feature
         valid_fraction_feature_type, valid_fraction_feature_name = self.valid_fraction_feature
 
         valid_frac = list(eopatch[valid_fraction_feature_type][valid_fraction_feature_name].flatten())
-        data = eopatch[feature_type][feature_name]
+        data = eopatch[feature_type, feature_name]
 
         number = data.shape[0] if self.number is None else self.number
 
-        eopatch[feature_type][new_feature_name] = np.array(
+        eopatch[feature_type, new_feature_name] = np.array(
             [data[x] for _, x in sorted(zip(valid_frac, range(data.shape[0])), reverse=True) if x <= number - 1]
         )
 
@@ -60,24 +66,23 @@ class BaseCompositingTask(EOTask, metaclass=ABCMeta):
     Contributor: Johannes Schmid, GeoVille Information Systems GmbH, 2018
 
     :param feature: Feature holding the input time-series. Default type is FeatureType.DATA
-    :type feature: (FeatureType, str)
     :param feature_composite: Type and name of output composite image. Default type is FeatureType.DATA_TIMELESS
-    :type feature_composite: (FeatureType, str)
     :param percentile: Percentile along the time dimension used for compositing. Methods use different percentiles
-    :type percentile: int or list
     :param max_index: Value used to flag indices with NaNs. Could be integer or NaN. Default is 255
-    :type max_index: int or NaN
     :param interpolation: Method used to compute percentile. Allowed values are {'geoville', 'linear', 'lower',
-                            'higher', 'midpoint', 'nearest'}. 'geoville' interpolation performs a custom
-                            implementation, while the other methods use the numpy `percentile` function. Default is
-                            'lower'
-    :type interpolation: str
+        'higher', 'midpoint', 'nearest'}. 'geoville' interpolation performs a custom implementation, while the other
+        methods use the numpy `percentile` function. Default is 'lower'.
     :param no_data_value: Value in the composite assigned to non-valid data points. Default is NaN
-    :type no_data_value: float or NaN
     """
 
     def __init__(
-        self, feature, feature_composite, percentile=None, max_index=255, interpolation="lower", no_data_value=np.nan
+        self,
+        feature: SingleFeatureSpec,
+        feature_composite: SingleFeatureSpec,
+        percentile: int,
+        max_index: int = 255,
+        interpolation: str = "lower",
+        no_data_value: float = np.nan,
     ):
         self.feature = self.parse_feature(feature)
         self.composite_type, self.composite_name = self.parse_feature(feature_composite)
@@ -91,13 +96,15 @@ class BaseCompositingTask(EOTask, metaclass=ABCMeta):
         )
         self.no_data_value = no_data_value
 
-    def _numpy_index_by_percentile(self, data, percentile):
+    def _numpy_index_by_percentile(self, data: np.ndarray, percentile: float) -> np.ndarray:
         """Calculate percentile of numpy stack and return the index of the chosen pixel.
 
         numpy percentile function is used with one of the following interpolations {'linear', 'lower', 'higher',
         'midpoint', 'nearest'}
         """
-        data_perc_low = np.nanpercentile(data, percentile, axis=0, interpolation=self.interpolation)
+        data_perc_low = np.nanpercentile(  # type: ignore[call-overload]  # numpy is weird here
+            data, percentile, axis=0, interpolation=self.interpolation
+        )
 
         indices = np.empty(data_perc_low.shape, dtype=np.uint8)
         indices[:] = np.nan
@@ -108,7 +115,7 @@ class BaseCompositingTask(EOTask, metaclass=ABCMeta):
 
         return indices
 
-    def _geoville_index_by_percentile(self, data, percentile):
+    def _geoville_index_by_percentile(self, data: np.ndarray, percentile: int) -> np.ndarray:
         """Calculate percentile of numpy stack and return the index of the chosen pixel."""
         # no_obs = bn.allnan(arr_tmp["data"], axis=0)
         data_tmp = np.array(data, copy=True)
@@ -126,30 +133,28 @@ class BaseCompositingTask(EOTask, metaclass=ABCMeta):
         # get floor value of reference band and index band
         ind = f_arr.astype("int16")
         y_val, x_val = ind_tmp.shape[1], ind_tmp.shape[2]
-        y_val, x_val = np.ogrid[0:y_val, 0:x_val]
+        y_val, x_val = np.ogrid[0:y_val, 0:x_val]  # type: ignore[assignment]
         idx = np.where(valid_obs == 0, self.max_index, ind_tmp[ind, y_val, x_val])
         return idx
 
-    def _get_reference_band(self, data):
+    @abstractmethod
+    def _get_reference_band(self, data: np.ndarray) -> np.ndarray:
         """Extract reference band from input 4D data according to compositing method
 
         :param data: 4D array from which to extract reference band (e.g. blue, maxNDVI, ..)
-        :type data: numpy array
         :return: 3D array containing reference band according to compositing method
         """
-        raise NotImplementedError
 
-    def _get_indices(self, data):
+    def _get_indices(self, data: np.ndarray) -> np.ndarray:
         """Compute indices along temporal dimension corresponding to the sought percentile
 
         :param data: Input 3D array holding the reference band
-        :type data: numpy array
         :return: 2D array holding the temporal index corresponding to percentile
         """
         indices = self._index_by_percentile(data, self.percentile)
         return indices
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """Compute composite array merging temporal frames according to the compositing method
 
         :param eopatch: eopatch holding time-series
@@ -169,7 +174,7 @@ class BaseCompositingTask(EOTask, metaclass=ABCMeta):
         for scene_id, scene in enumerate(data):
             composite_image = np.where(np.dstack([indices]) == scene_id, scene, composite_image)
 
-        eopatch[self.composite_type][self.composite_name] = composite_image
+        eopatch[self.composite_type, self.composite_name] = composite_image
 
         return eopatch
 
@@ -180,20 +185,24 @@ class BlueCompositingTask(BaseCompositingTask):
     - blue     (25th percentile of the blue band)
 
     :param blue_idx: Index of blue band in `feature` array
-    :type blue_idx: int
     """
 
-    def __init__(self, feature, feature_composite, blue_idx, interpolation="lower"):
+    def __init__(
+        self,
+        feature: SingleFeatureSpec,
+        feature_composite: SingleFeatureSpec,
+        blue_idx: int,
+        interpolation: str = "lower",
+    ):
         super().__init__(feature, feature_composite, percentile=25, interpolation=interpolation)
         self.blue_idx = blue_idx
         if not isinstance(blue_idx, int):
             raise ValueError("Incorrect value of blue band index specified")
 
-    def _get_reference_band(self, data):
+    def _get_reference_band(self, data: np.ndarray) -> np.ndarray:
         """Extract the blue band from time-series
 
         :param data: 4D array from which to extract the blue reference band
-        :type data: numpy array
         :return: 3D array containing the blue reference band
         """
         return data[..., self.blue_idx].astype("float32")
@@ -209,23 +218,27 @@ class HOTCompositingTask(BaseCompositingTask):
             Remote Sensing of Environment, 118, 83-94.
 
     :param blue_idx: Index of blue band in `feature` array
-    :type blue_idx: int
     :param red_idx: Index of red band in `feature` array
-    :type red_idx: int
     """
 
-    def __init__(self, feature, feature_composite, blue_idx, red_idx, interpolation="lower"):
+    def __init__(
+        self,
+        feature: SingleFeatureSpec,
+        feature_composite: SingleFeatureSpec,
+        blue_idx: int,
+        red_idx: int,
+        interpolation: str = "lower",
+    ):
         super().__init__(feature, feature_composite, percentile=25, interpolation=interpolation)
         self.blue_idx = blue_idx
         self.red_idx = red_idx
         if not isinstance(blue_idx, int) or not isinstance(red_idx, int):
             raise ValueError("Incorrect values of blue and red band indices specified")
 
-    def _get_reference_band(self, data):
+    def _get_reference_band(self, data: np.ndarray) -> np.ndarray:
         """Extract the HOT band from time-series
 
         :param data: 4D array from which to extract the HOT reference band
-        :type data: numpy array
         :return: 3D array containing the HOT reference band
         """
         return data[..., self.blue_idx] - 0.5 * data[..., self.red_idx] - 0.08
@@ -237,33 +250,40 @@ class MaxNDVICompositingTask(BaseCompositingTask):
     - maxNDVI  (temporal maximum of NDVI)
 
     :param red_idx: Index of red band in `feature` array
-    :type red_idx: int
     :param nir_idx: Index of NIR band in `feature` array
-    :type nir_idx: int
     """
 
-    def __init__(self, feature, feature_composite, red_idx, nir_idx, interpolation="lower"):
-        super().__init__(feature, feature_composite, percentile=[0, 100], interpolation=interpolation)
+    def __init__(
+        self,
+        feature: SingleFeatureSpec,
+        feature_composite: SingleFeatureSpec,
+        red_idx: int,
+        nir_idx: int,
+        interpolation: str = "lower",
+    ):
+        super().__init__(  # No real value for percentile
+            feature, feature_composite, percentile=None, interpolation=interpolation  # type: ignore[arg-type]
+        )
+        self.percentiles = [0, 100]
         self.red_idx = red_idx
         self.nir_idx = nir_idx
         if not isinstance(nir_idx, int) or not isinstance(red_idx, int):
             raise ValueError("Incorrect values of red and NIR band indices specified")
 
-    def _get_reference_band(self, data):
+    def _get_reference_band(self, data: np.ndarray) -> np.ndarray:
         """Extract the NDVI band from time-series
 
         :param data: 4D array from which to compute the NDVI reference band
-        :type data: numpy array
         :return: 3D array containing the NDVI reference band
         """
         nir = data[..., self.nir_idx].astype("float32")
         red = data[..., self.red_idx].astype("float32")
         return (nir - red) / (nir + red)
 
-    def _get_indices(self, data):
+    def _get_indices(self, data: np.ndarray) -> np.ndarray:
         median = np.nanmedian(data, axis=0)
-        indices_min = self._index_by_percentile(data, self.percentile[0])
-        indices_max = self._index_by_percentile(data, self.percentile[1])
+        indices_min = self._index_by_percentile(data, self.percentiles[0])
+        indices_max = self._index_by_percentile(data, self.percentiles[1])
         indices = np.where(median < -0.05, indices_min, indices_max)
         return indices
 
@@ -274,23 +294,27 @@ class MaxNDWICompositingTask(BaseCompositingTask):
     - maxNDWI  (temporal maximum of NDWI)
 
     :param nir_idx: Index of NIR band in `feature` array
-    :type nir_idx: int
     :param swir1_idx: Index of SWIR1 band in `feature` array
-    :type swir1_idx: int
     """
 
-    def __init__(self, feature, feature_composite, nir_idx, swir1_idx, interpolation="lower"):
+    def __init__(
+        self,
+        feature: SingleFeatureSpec,
+        feature_composite: SingleFeatureSpec,
+        nir_idx: int,
+        swir1_idx: int,
+        interpolation: str = "lower",
+    ):
         super().__init__(feature, feature_composite, percentile=100, interpolation=interpolation)
         self.nir_idx = nir_idx
         self.swir1_idx = swir1_idx
         if not isinstance(nir_idx, int) or not isinstance(swir1_idx, int):
             raise ValueError("Incorrect values of NIR and SWIR1 band indices specified")
 
-    def _get_reference_band(self, data):
+    def _get_reference_band(self, data: np.ndarray) -> np.ndarray:
         """Extract the NDWI band from time-series
 
         :param data: 4D array from which to compute the NDWI reference band
-        :type data: numpy array
         :return: 3D array containing the NDWI reference band
         """
         nir = data[..., self.nir_idx].astype("float32")
@@ -304,14 +328,19 @@ class MaxRatioCompositingTask(BaseCompositingTask):
     - maxRatio (temporal maximum of a ratio using bands blue, NIR and SWIR)
 
     :param blue_idx: Index of blue band in `feature` array
-    :type blue_idx: int
     :param nir_idx: Index of NIR band in `feature` array
-    :type nir_idx: int
     :param swir1_idx: Index of SWIR1 band in `feature` array
-    :type swir1_idx: int
     """
 
-    def __init__(self, feature, feature_composite, blue_idx, nir_idx, swir1_idx, interpolation="lower"):
+    def __init__(
+        self,
+        feature: SingleFeatureSpec,
+        feature_composite: SingleFeatureSpec,
+        blue_idx: int,
+        nir_idx: int,
+        swir1_idx: int,
+        interpolation: str = "lower",
+    ):
         super().__init__(feature, feature_composite, percentile=100, interpolation=interpolation)
         self.blue_idx = blue_idx
         self.nir_idx = nir_idx
@@ -319,13 +348,12 @@ class MaxRatioCompositingTask(BaseCompositingTask):
         if not isinstance(blue_idx, int) or not isinstance(nir_idx, int) or not isinstance(swir1_idx, int):
             raise ValueError("Incorrect values for either blue, NIR or SWIR1 band indices specified")
 
-    def _get_reference_band(self, data):
+    def _get_reference_band(self, data: np.ndarray) -> np.ndarray:
         """Extract the max-ratio band from time-series
 
         The max-ratio is defined as max(NIR,SWIR1)/BLUE
 
         :param data: 4D array from which to compute the max-ratio reference band
-        :type data: numpy array
         :return: 3D array containing the max-ratio reference band
         """
         blue = data[..., self.blue_idx].astype("float32")
@@ -341,34 +369,31 @@ class HistogramMatchingTask(EOTask):
     Contributor: Johannes Schmid, GeoVille Information Systems GmbH, 2018
 
     :param feature: Name of the eopatch data layer that will undergo a histogram match.
-                    Should be of the FeatureType "DATA".
-    :type feature: (FeatureType, str) or (FeatureType, str, str)
+        Should be of the FeatureType "DATA".
     :param reference: Name of the eopatch data layer that represents the reference for the histogram match.
-                        Should be of the FeatureType "DATA_TIMELESS".
-    :type reference: (FeatureType, str)
+        Should be of the FeatureType "DATA_TIMELESS".
     """
 
-    def __init__(self, feature, reference):
+    def __init__(self, feature: SingleFeatureSpec, reference: SingleFeatureSpec):
         self.renamed_feature = self.parse_renamed_feature(feature)
         self.reference = self.parse_feature(reference)
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """Perform histogram matching of the time-series with respect to a reference scene
 
         :param eopatch: eopatch holding the time-series and reference data
-        :type eopatch: EOPatch
         :return: The same eopatch instance with the normalised time-series
         """
         feature_type, feature_name, new_feature_name = self.renamed_feature
         reference_type, reference_name = self.reference
 
-        reference_scene = eopatch[reference_type][reference_name]
+        reference_scene = eopatch[reference_type, reference_name]
         # check if band dimension matches
-        if reference_scene.shape[-1] != eopatch[feature_type][feature_name].shape[-1]:
+        if reference_scene.shape[-1] != eopatch[feature_type, feature_name].shape[-1]:
             raise ValueError("Time-series and reference scene must have corresponding bands")
 
-        eopatch[feature_type][new_feature_name] = np.zeros_like(eopatch[feature_type][feature_name])
-        for source_id, source in enumerate(eopatch[feature_type][feature_name]):
+        eopatch[feature_type, new_feature_name] = np.zeros_like(eopatch[feature_type, feature_name])
+        for source_id, source in enumerate(eopatch[feature_type, feature_name]):
             # mask-out same invalid pixels
             src_masked = np.where(np.isnan(reference_scene), np.nan, source)
             ref_masked = np.where(np.isnan(source), np.nan, reference_scene)
@@ -378,7 +403,7 @@ class HistogramMatchingTask(EOTask):
             mean_ref = np.nanmean(ref_masked, axis=(0, 1), dtype=np.float64)
             mean_src = np.nanmean(src_masked, axis=(0, 1), dtype=np.float64)
             # normalise values
-            eopatch[feature_type][new_feature_name][source_id] = source * (std_ref / std_src) + (
+            eopatch[feature_type, new_feature_name][source_id] = source * (std_ref / std_src) + (
                 mean_ref - (mean_src * (std_ref / std_src))
             )
 
